import os

import platform
import unittest
import numpy as np
import tensorflow as tf
from yaket.schema.schema import Accelerator, TrainingModel
from yaket.trainer import Trainer
import psutil

class CModel(tf.keras.Model):
    """Simple custom Model to test Trainer"""

    def __init__(self):
        super(CModel, self).__init__()
        self.dense = tf.keras.layers.Dense(1)

    def call(self, inputs):
        return self.dense(inputs)


class CLayer(tf.keras.layers.Layer):
    """Simple custom Layer to test Trainer"""

    def __init__(self):
        super(CLayer, self).__init__()
        self.dense = tf.keras.layers.Dense(1)

    def call(self, inputs):
        return self.dense(inputs)


class TestTrainer(unittest.TestCase):

    optimizers = Trainer.list_available_tf_modules("optimizers")
    losses = Trainer.list_available_tf_modules("losses")
    metrics = Trainer.list_available_tf_modules("metrics")

    simple_model = tf.keras.Sequential()
    simple_model.add(tf.keras.layers.Dense(1, input_shape=(1,)))

    train_dataset = tf.data.Dataset.from_tensor_slices((range(10), range(10))).batch(1)
    dev_dataset = tf.data.Dataset.from_tensor_slices((range(10), range(10))).batch(1)

    simple_config = dict(
        optimizer="Adam",
        loss="BinaryCrossentropy",
        metrics=["Accuracy"],
        epochs=1,
        verbose=2,
        shuffle=True,
        accelerator="cpu",
    )

    boring_trainer = Trainer(
        simple_config,
        simple_model,
        train_dataset=train_dataset,
        val_dataset=dev_dataset,
    )



    def test_create_new_trainer(self):
        """
        GIVEN a Trainer initialized with valid config params, model, and train/dev dataset
        WHEN a new Trainer instance is created
        THEN a new Trainer instance exists with same attributes
        """

        trainer = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=self.train_dataset,
            val_dataset=self.dev_dataset,
        )

        assert trainer.model == self.simple_model
        assert trainer.train_dataset == self.train_dataset
        assert trainer.val_dataset == self.dev_dataset
        assert isinstance(trainer.config, TrainingModel)

    def test_train_history(self):
        """Test whether training is working.
        GIVEN a correct initialized Trainer
        WHEN train() is called
        THEN test the train history exists
        """

        trainer = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=self.train_dataset,
            val_dataset=self.dev_dataset,
        )
        history = trainer.train()
        assert history is not None

    def test_train_saved(self):
        """Test whether training is working and model is saved
        GIVEN a correct initialized Trainer
        WHEN train() is called
        THEN test the model is saved
        """

        trainer = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=self.train_dataset,
            val_dataset=self.dev_dataset,
        )
        trainer.train()
        assert os.path.exists(trainer._out_path)

    def test_convert_model_with_no_trained_model(self):
        """Test whether convert_model() is called without a trained model
        GIVEN a correct initialized Trainer
        WHEN convert_model() is called
        THEN an error is raised if the model has not been saved"""

        with self.assertRaises(Exception):
            self.boring_trainer.convert_model("onnx")

    def test_convert_model_wrong_format_model(self):
        """GIVEN a wrong format model
        WHEN convert_model() is called
        THEN an error is raised
        """

        format_list = [2,'random_string', (1,2), [1,2]]

        trainer = Trainer(
                    self.simple_config,
                    self.simple_model,
                    train_dataset=self.train_dataset,
                    val_dataset=self.dev_dataset,
                )
        trainer.train()

        for format_type in format_list:
            with self.subTest(msg=format_type):
                with self.assertRaises(Exception):
                    trainer.convert_model(format_type)

    def test_convert_model_wrong_opset_onnx(self):
        """GIVEN a wrong opset onnx
        WHEN convert_model() is called
        THEN an error is raised
        """

        opset_list = ['random_string', -1, [1,2]]

        trainer = Trainer(
                    self.simple_config,
                    self.simple_model,
                    train_dataset=self.train_dataset,
                    val_dataset=self.dev_dataset,
                )
        trainer.train()

        for opset in opset_list:
            with self.subTest(msg=opset):
                with self.assertRaises(TypeError):
                    trainer.convert_model(opset_onnx = opset)

    


    def test_init_wrong_model_format(self):
        """GIVEN a wrong format model
        WHEN Trainer is initialized
        THEN an error is raised
        """

        model_list = [2,'random_string', (1,2), [1,2]]

        for model in model_list:
            with self.subTest(msg=model):
                with self.assertRaises(Exception):
                    Trainer(
                        self.simple_config,
                        model,
                        train_dataset=self.train_dataset,
                        val_dataset=self.dev_dataset,
                    )

    def test_init_long_tuple_train_dataset(self):
        """GIVEN a train dataset with more than 3 elements
        WHEN Trainer is initialized
        THEN an error is raised
        """

        train_dataset = tf.data.Dataset.from_tensor_slices((range(10), range(10), range(10), range(10))).batch(1)

        with self.assertRaises(Exception):
            Trainer(
                self.simple_config,
                self.simple_model,
                train_dataset=train_dataset,
                val_dataset=self.dev_dataset,
            )
        
    
    def test_init_list_train_dataset(self):
        """GIVEN a train dataset with a list
        WHEN Trainer is initialized
        THEN an error is raised
        """

        train_dataset = [range(10), range(10)]

        with self.assertRaises(Exception):
            Trainer(
                self.simple_config,
                self.simple_model,
                train_dataset=train_dataset,
                val_dataset=self.dev_dataset,
            )
    def test_init_list_dev_dataset(self):
        """GIVEN a dev dataset with a list
        WHEN Trainer is initialized
        THEN an error is raised
        """

        dev_dataset = 'string'

        
        with self.assertRaises(TypeError):
            Trainer(
                self.simple_config,
                self.simple_model,
                train_dataset=self.train_dataset,
                val_dataset=dev_dataset,
            )

    def test_init_wrong_strategy_format(self):
        """GIVEN a wrong format strategy
        WHEN Trainer is initialized
        THEN an error is raised
        """

        strategy_list = [2,'random_string', (1,2), [1,2]]

        for strategy in strategy_list:
            with self.subTest(msg=strategy):
                with self.assertRaises(Exception):
                    Trainer(
                        self.simple_config,
                        self.simple_model,
                        train_dataset=self.train_dataset,
                        val_dataset=self.dev_dataset,
                        strategy=strategy
                    )
            
    def test_init_validate_yaml_wrong_format(self):
        """GIVEN a wrong format yaml file
        WHEN Trainer is initialized
        THEN an error is raised
        """

        validate_yaml_list = [2,'random_string', (1,2), [1,2]]

        for yaml in validate_yaml_list:
            with self.subTest(msg=yaml):
                with self.assertRaises(Exception):
                    Trainer(
                        self.simple_config,
                        self.simple_model,
                        train_dataset=self.train_dataset,
                        val_dataset=self.dev_dataset,
                        validate_yaml=yaml
                    )
    def test_init_random_seed_wrong_format(self):
        """GIVEN a wrong format random seed
        WHEN Trainer is initialized
        THEN an error is raised
        """

        random_seed_list = ['random_string', (1,2), [1,2]]

        for random_seed in random_seed_list:
            with self.subTest(msg=random_seed):
                with self.assertRaises(Exception):
                    Trainer(
                        self.simple_config,
                        self.simple_model,
                        train_dataset=self.train_dataset,
                        val_dataset=self.dev_dataset,
                        random_seed=random_seed
                    )

    

    def test_convert_model_wrong_output_path(self):
        """GIVEN a wrong output path
        WHEN convert_model() is called
        THEN an error is raised
        """

        output_path_list = [2, 1.2, [1,2]]

        trainer = Trainer(
                    self.simple_config,
                    self.simple_model,
                    train_dataset=self.train_dataset,
                    val_dataset=self.dev_dataset,
                )
        trainer.train()

        for output_path in output_path_list:
            with self.subTest(msg=output_path):
                with self.assertRaises(TypeError):
                    trainer.convert_model(output_path = output_path)
    
    def test_convert_model_wrong_from_command_line(self):
        """GIVEN a wrong from command line
        WHEN convert_model() is called
        THEN an error is raised
        """

        from_command_line_list = [2, 1.2, [1,2],'random_string']

        trainer = Trainer(
                    self.simple_config,
                    self.simple_model,
                    train_dataset=self.train_dataset,
                    val_dataset=self.dev_dataset,
                )
        trainer.train()

        for from_command_line in from_command_line_list:
            with self.subTest(msg=from_command_line):
                with self.assertRaises(TypeError):
                    trainer.convert_model(from_command_line = from_command_line)

    def test_get_sample_weight_mode_none(self):
        """Test whether get_sample_weight_mode() is called
        GIVEN a correct initialized Trainer without sample_weight_mode
        WHEN get_sample_weight_mode() is called
        THEN test the sample weight mode is None
        """

        trainer = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=self.train_dataset,
            val_dataset=self.dev_dataset,
        )
        sample_weight_mode = trainer._get_sample_weight_mode()
        self.assertIsNone(sample_weight_mode)
    
    def test_get_sample_weight_mode_temporal(self):
        """GIVEN a correct initialized Trainer with sample_weight_mode = 'temporal'
        WHEN get_sample_weight_mode() is called
        THEN test the sample weight mode is 'temporal'"""
        d = self.simple_config
        d['sample_weight_mode'] = 'temporal'
        trainer = Trainer(d,
            self.simple_model,
            train_dataset=self.train_dataset,
            val_dataset=self.dev_dataset,)
        sample_weight_mode = trainer._get_sample_weight_mode()
        self.assertEqual(sample_weight_mode, 'temporal')

    def test_get_input_shape_tf_data(self):
        """GIVEN a correct initialized Trainer with tf.data.Dataset
        WHEN get_input_shape() is called
        THEN test the input shape is correct
        """


        train_dataset_numpy= (np.random.random((2,10)), np.random.random((2,10)))
        dev_dataset_numpy = (np.random.random((2,10)), np.random.random((2,10)))

        train_dataset = tf.data.Dataset.from_tensor_slices(train_dataset_numpy).batch(1)
        dev_dataset = tf.data.Dataset.from_tensor_slices(dev_dataset_numpy).batch(1)

        expected_input_shape = (None, 10)

        trainer = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=train_dataset,
            val_dataset=dev_dataset,
        )
        input_shape = trainer._get_input_shape()
        np.testing.assert_array_equal(input_shape, expected_input_shape)

    def test_get_x_y_train_output_type(self):
        """GIVEN a correct initialized Trainer with tf.data.Dataset or numpy array
        WHEN get_x_y_val() is called
        THEN test the output type is tf.data.Dataset
        """

        train_dataset_numpy= (np.random.random((2,10)), np.random.random((2,10)))
        

        trainer_numpy = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=train_dataset_numpy,
        )
        trainer_tf = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=self.train_dataset,
        )

        x_np = trainer_numpy._get_x_y_train(1)
        x_tf = trainer_tf._get_x_y_train(1)
        self.assertIsInstance(x_np, tf.data.Dataset)
        self.assertIsInstance(x_tf, tf.data.Dataset)

    def test_get_x_y_val_output_type(self):
        """GIVEN a correct initialized Trainer with tf.data.Dataset, numpy array, or None
        WHEN get_x_y_val() is called
        THEN test the output type is tf.data.Dataset or None if val_dataset is None
        """

        dev_dataset_numpy= (np.random.random((2,10)), np.random.random((2,10)))
        

        trainer_numpy = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=self.train_dataset,
            val_dataset=dev_dataset_numpy,
        )
        trainer_tf = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=self.train_dataset,
            val_dataset=self.dev_dataset,
        )
        trainer_none = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=self.train_dataset,
        )


        x_np = trainer_numpy._get_x_y_val(1)
        x_tf = trainer_tf._get_x_y_val(1)
        x_none = trainer_none._get_x_y_val(1)
        self.assertIsInstance(x_np, tf.data.Dataset)
        self.assertIsInstance(x_tf, tf.data.Dataset)
        self.assertIsNone(x_none)

    def test_validate_config(self):
        """GIVEN a correct initialized Trainer
        WHEN a config parameter is changed with a wrong type and validate_config() is called
        THEN test the config is validated by throwing an error
        """

        trainer = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=self.train_dataset,
            val_dataset=self.dev_dataset,
        )

        trainer.config.epochs = 'wrong_type'
        with self.assertRaises(Exception):
            trainer.validate_config()
    
    def test_validate_config_no_error(self):
        """GIVEN a correct initialized Trainer
        WHEN a config parameter is changed with the correct type and validate_config() is called
        THEN test the config is validated without throwing an error and the new value is set
        """

        new_value_config_epochs = 10

        trainer = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=self.train_dataset,
            val_dataset=self.dev_dataset,
        )

        trainer.config.epochs = 10

        trainer.validate_config()
        self.assertEqual(trainer.config.epochs, new_value_config_epochs)
    

    
    def test_get_input_shape_numpy_data(self):
        """GIVEN a correct initialized Trainer with numpy array
        WHEN get_input_shape() is called
        THEN test the input shape is correct
        """

        train_dataset_numpy= (np.random.random((2,10)), np.random.random((2,10)))
        dev_dataset_numpy = (np.random.random((2,10)), np.random.random((2,10)))

        expected_input_shape = (None, 10)

        trainer = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=train_dataset_numpy,
            val_dataset=dev_dataset_numpy,
        )
        input_shape = trainer._get_input_shape()
        np.testing.assert_array_equal(input_shape, expected_input_shape)

    def test_get_free_gpu_idx_linux_no_gpu(self):
        """GIVEN a Trainer class
        WHEN get_free_gpu_idx() is called on a linux machine without GPU 
        THEN test an error is thrown
        """
        with self.assertRaises(Exception):
            Trainer.get_free_gpu_idx()
    
    def test_get_free_gpu_idx_gpu(self):
        """GIVEN a Trainer class
        WHEN get_free_gpu_idx() is called on a linux machine with GPU 
        THEN test the first GPU is returned
        """
        gpus = tf.config.experimental.list_physical_devices("GPU")
        if gpus:
            self.assertEqual(Trainer.get_free_gpu_idx(), 0)
        
    def test_get_free_gpu_idx_windows(self):
        """GIVEN a Trainer class
        WHEN get_free_gpu_idx() is called on a windows machine 
        THEN test an error is thrown
        """
        if platform.system == 'Windows':
            with self.assertRaises(Exception):
                Trainer.get_free_gpu_idx()

    def test_list_available_tf_modules_type_error(self):
        """GIVEN a Trainer class
        WHEN list_available_tf_modules() is called with a wrong type
        THEN an error is thrown
        """

        random_options = ['callabacks','tf_addons','netric']
        for option in random_options:
            with self.assertRaises(AssertionError):
                Trainer.list_available_tf_modules(option)
    
        
    
    def test_set_randomness_seed_wrong(self):
        """GIVEN a Trainer class
        WHEN _set_randomness() is called with a wrong seed
        THEN test an error is thrown
        """
        seeds = ['wrong_type', -1,[1,2]]
        for seed in seeds:
            with self.subTest(seed=seed):
                with self.assertRaises(ValueError):
                    self.boring_trainer._set_randomness(seed)

    def test_clear_ram(self):
        """GIVEN a Trainer class correctly initialized and trained
        WHEN clear_ram() is called
        THEN test the RAM is cleared
        """

        trainer = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=self.train_dataset,
        )
        
        history = trainer.train(100)

        ram_before_clearing = psutil.virtual_memory().percent

        trainer.clear_ram()
        ram_after_clearing = psutil.virtual_memory().percent

        self.assertGreaterEqual(ram_before_clearing, ram_after_clearing)

    def test_summary_model_no_initialized(self):
        """GIVEN a Trainer class
        WHEN summary_model() is called before the model is initialized
        THEN test an error is thrown
        """
        with self.assertRaises(Exception):
            Trainer.summary_model()


    def test_wrong_type_config_params(self):
        """GIVEN a Trainer class 
        WHEN it is initialized  with a wrong param_config type
        THEN test an error is thrown
        """

        params = [2, (3.0, 2), list([1, 2])]

        for param in params:
            with self.subTest(msg=param):
                with self.assertRaises(TypeError):
                    Trainer(
                        param,
                        self.simple_model,
                        train_dataset=self.train_dataset,
                        val_dataset=self.dev_dataset,
                    )

    def test_wrong_values_config_params(self):
        """Test whether a not-existing file has been passed or a wrong-type dictionary
        GIVEN a Trainer class
        WHEN it is initialized with a wrong param_config value
        THEN test an error is thrown
        """

        params = ["random_path.yaml", dict(a=1, b=2)]

        with self.subTest():
            with self.assertRaises(FileNotFoundError):
                Trainer(
                    params[0],
                    self.simple_model,
                    train_dataset=self.train_dataset,
                    val_dataset=self.dev_dataset,
                )
        with self.subTest():
            with self.assertRaises(ValueError):
                Trainer(
                    params[1],
                    self.simple_model,
                    train_dataset=self.train_dataset,
                    val_dataset=self.dev_dataset,
                )

    def test_wrong_keras_optimizer(self):
        """Test whether a wrong keras optimizer has been passed.
        GIVEN a Trainer class
        WHEN it is initialized with a wrong keras optimizer
        THEN test an error is thrown
        """

        params = ["random_optimizer", "adam", "AdamW"]

        for param in params:
            with self.subTest(msg=param):
                with self.assertRaises(ValueError):
                    d = self.simple_config
                    d["optimizer"] = param
                    Trainer(
                        d,
                        self.simple_model,
                        train_dataset=self.train_dataset,
                        val_dataset=self.dev_dataset,
                    )

    def test_wrong_keras_loss(self):
        """Test whether a wrong keras loss has been passed.
        GIVEN a Trainer class
        WHEN it is initialized with a wrong keras loss
        THEN test an error is thrown
        """

        params = ["random_loss", "binary_crossentropy", "mse"]

        for param in params:
            with self.subTest(msg=param):
                with self.assertRaises(ValueError):
                    d = self.simple_config
                    d["loss"] = param
                    Trainer(
                        d,
                        self.simple_model,
                        train_dataset=self.train_dataset,
                        val_dataset=self.dev_dataset,
                    )

    def test_wrong_keras_metric(self):
        """Test whether a wrong keras metric has been passed.
        GIVEN a Trainer class
        WHEN it is initialized with a wrong keras metric
        THEN test an error is thrown
        """

        params = ["prc", "truepositive", "mse"]

        for param in params:
            with self.subTest(msg=param):
                with self.assertRaises(ValueError):
                    d = self.simple_config
                    d["metrics"] = param
                    Trainer(
                        d,
                        self.simple_model,
                        train_dataset=self.train_dataset,
                        val_dataset=self.dev_dataset,
                    )

    def test_import_custom_module(self):
        """Test whether a wrong custom module can be loaded.
        GIVEN a Trainer class
        WHEN it is initialized with a wrong custom module
        THEN test an error is thrown
        """

        custom_module = "custom_module.py"
        with self.assertRaises(FileNotFoundError):
            Trainer(
                self.simple_config,
                self.simple_model,
                train_dataset=self.train_dataset,
                val_dataset=self.dev_dataset,
                custom_modules_path=custom_module,
            )


    def test_accelerator(self):
        """Test whether accelerator config param is converted to right Enum member.
        Test only for CPU because it throws error if GPU is not found
        GIVEN a Trainer class
        WHEN it is initialized with a wrong accelerator
        THEN test an error is thrown"""

        params = ["cpu"]
        acceleator_expected = [
            Accelerator.cpu,
            Accelerator.gpu,
            Accelerator.mgpu,
            Accelerator.tpu,
        ]
        for param, expected in zip(params, acceleator_expected):
            with self.subTest(msg=param):
                d = self.simple_config
                d["accelerator"] = param
                trainer = Trainer(
                    d,
                    self.simple_model,
                    train_dataset=self.train_dataset,
                    val_dataset=self.dev_dataset,
                )
                self.assertEqual(trainer._accelerator, expected)

    def test_default_strategy(self):
        """Test whether strategy is correctly initialized.
        GIVEN a Trainer class
        WHEN it is initialized with a wrong accelerator
        THEN test an error is thrown

        TODO: tpu is not implemented yet"""

        params = ["cpu"]
        expected_strategies = [tf.distribute.MirroredStrategy]
        strategy = None

        if tf.config.list_physical_devices("GPU"):
            params.extend(["gpu", "mgpu"])
            expected_strategies.extend(
                [tf.distribute.OneDeviceStrategy, tf.distribute.MirroredStrategy]
            )

        for param, expected in zip(params, expected_strategies):
            with self.subTest(msg=param):
                d = self.simple_config
                d["accelerator"] = param
                trainer = Trainer(
                    d,
                    self.simple_model,
                    train_dataset=self.train_dataset,
                    val_dataset=self.dev_dataset,
                    strategy=strategy,
                )

                self.assertIsNone(trainer.strategy)

                strategy_out = trainer._get_strategy()
                self.assertIsInstance(strategy_out, expected)

    def test_compile_model(self):
        """Test whether model is compiled correctly.
        GIVEN a correctly initialized Trainer class
        WHEN it _compile_model is called
        THEN test the model is compiled

        """

        params = ["cpu"]
        expected_strategies = [tf.distribute.MirroredStrategy]
        strategy = None

        if tf.config.list_physical_devices("GPU"):
            params.extend(["gpu", "mgpu"])
            expected_strategies.extend(
                [tf.distribute.OneDeviceStrategy, tf.distribute.MirroredStrategy]
            )

        for param, expected in zip(params, expected_strategies):
            with self.subTest(msg=param):
                d = self.simple_config
                d["accelerator"] = param
                trainer = Trainer(
                    d,
                    self.simple_model,
                    train_dataset=self.train_dataset,
                    val_dataset=self.dev_dataset,
                    strategy=strategy,
                )

                trainer._compile_model()
                self.assertTrue(trainer.model._is_compiled)

    def test_custom_model_clone_model(self):
        """Test whether custom model is not cloned.
        GIVEN a correctly initialized Trainer class with a custom model
        WHEN it _clone_model is called
        THEN test the model is not cloned and error is thrown
        """

        model = CModel()
        trainer = Trainer(
            self.simple_config,
            model,
            train_dataset=self.train_dataset,
            val_dataset=self.dev_dataset,
        )
        with self.assertRaises(Exception):
            trainer._clone_model()

    def test_custom_layer_clone_model(self):
        """Test whether model with custom layer is not cloned correctly.
        GIVEN a correctly initialized Trainer class with a custom layer
        WHEN it _clone_model is called
        THEN test the model is not cloned and error is thrown"""

        model = tf.keras.Sequential([CLayer()])
        trainer = Trainer(
            self.simple_config,
            model,
            train_dataset=self.train_dataset,
            val_dataset=self.dev_dataset,
        )
        with self.assertRaises(Exception):
            trainer._clone_model()

    def test_weights_save_model(self):
        """Test whether vanilla model is saved with weights and it can be loaded back
        
        GIVEN a correctly initialized Trainer class with a vanilla model
        WHEN it _save_model is called
        THEN test the model is saved with weights and it can be loaded back
        """

        model = CModel()
        trainer = Trainer(
            self.simple_config,
            model,
            train_dataset=self.train_dataset,
            val_dataset=self.dev_dataset,
        )

        trainer._save_model()

        self.assertTrue(os.path.exists(trainer._out_path + ".index"))
        trainer.model.load_weights(trainer._out_path)

    def test_savedmodel_save_model(self):
        """Test whether vanilla model is saved in SavedFormat and it can be loaded back
       
        GIVEN a correctly initialized Trainer class with a vanilla model
        WHEN it _save_model is called
        THEN test the model is saved in SavedFormat and it can be loaded back

        """

        trainer = Trainer(
            self.simple_config,
            self.simple_model,
            train_dataset=self.train_dataset,
            val_dataset=self.dev_dataset,
        )

        trainer._save_model()

        self.assertTrue(os.path.exists(trainer._out_path))
        tf.keras.models.load_model(trainer._out_path)


if __name__ == "__main__":

    unittest.main()



import os

import numpy as np
import tensorflow as tf
from yaket.schema.schema import Accelerator, TrainingModel
from yaket.trainer import Trainer
from yaket.converter.converter import Converter


class TestConverter(tf.test.TestCase):

    boring_model = tf.keras.Sequential(
        [
        tf.keras.layers.Dense(10, activation='relu', input_shape=(32,)),
        tf.keras.layers.Dense(10, activation='relu'),
        tf.keras.layers.Dense(1, activation='sigmoid')
        ]           
    )

    def test_init_converter_wrong_out_path(self):
        """
        GIVEN an out_path invalid parameter
        WHEN the converter is initialized
        THEN the converter should raise an exception
        """ 
        out_format = 'onnx'
        out_paths = [[1,2], (3,2), 1]

        for out_path in out_paths:
            with self.subTest(out_path=out_path):
                with self.assertRaises(TypeError):
                    Converter(out_format=out_format, out_path=out_path)
    
        
    def test_init_converter_wrong_out_format(self):
        """
        GIVEN an out_format invalid parameter
        WHEN the _init_converter method is called
        THEN the converter should raise an exception
        """ 
        out_format_list = [ 1, 1.0, ['string']]
        
        for out_format in out_format_list:
            with self.subTest(out_format=out_format):
                with self.assertRaises(TypeError):
                    c = Converter(out_format=out_format)

    

    def test_init_convert_wrong_type_onnx(self):
        """
        GIVEN a Converter correctly initialized with onnx out_format
        WHEN the convert method is called with a wrong type
        THEN the model should raise an exception
        """ 
        out_format = 'onnx'
        opset_onnx = 'ciao'
        
        with self.assertRaises(TypeError):
            Converter(out_format=out_format, opset_onnx=opset_onnx)

    def test_init_converter_wrong_opset_onnx(self):
        """
        GIVEN an opset_onnx invalid parameter
        WHEN the converter is initialized
        THEN the converter should raise an exception
        """ 
        out_format = 'onnx'
        opset_onnx = 0
        
        with self.assertRaises(ValueError):
            Converter(out_format=out_format, opset_onnx=opset_onnx)

    def test_init_converter_wrong_model(self):
        """
        GIVEN a model invalid parameter
        WHEN the converter is initialized
        THEN the converter should raise an exception
        """ 
        out_format = 'onnx'
        model_list = [[1,2], (3,2), 1, 'string']

        for model in model_list:
            with self.subTest(model=model):
                with self.assertRaises(TypeError):
                    Converter(out_format=out_format, model=model)

    def test_init_converter_wrong_model_path(self):
        """
        GIVEN a model_path invalid parameter
        WHEN the converter is initialized
        THEN the converter should raise an exception
        """ 
        out_format = 'onnx'
        model_paths = [[1,2], (3,2), 1]

        for model_path in model_paths:
            with self.subTest(model_path=model_path):
                with self.assertRaises(TypeError):
                    Converter(out_format=out_format, model_path=model_path)
            

    def test_init_converter_default(self):
        """
        GIVEN an out_format valid parameter
        WHEN the converter is initialized
        THEN the converter should be initialized with the correct default attributes
        """ 
        out_format = 'onnx'
        
        converter = Converter(out_format=out_format)
    

        self.assertEqual(converter.out_format, out_format)
        self.assertIsNone(converter.model)
        self.assertIsNone(converter.model_path)
        self.assertEqual(converter.out_path, 'model.onnx')
        self.assertEqual(converter.opset_onnx, 15)

    def test_convert_onnx_format(self):
        """
        GIVEN a Converter correctly initialized with onnx out_format
        WHEN the convert method is called
        THEN the model should be converted to onnx format
        """ 
        out_format = 'onnx'
        converter = Converter(out_format=out_format, model=self.boring_model)
        converter.convert()
        self.assertTrue(os.path.exists(converter.out_path))


    def test_convert_onnx_format_with_cli(self):
        """
        GIVEN a Converter correctly initialized with onnx out_format and without model
        WHEN the convert method is called with the cli
        THEN the model should be converted to onnx format
        """ 
        out_format = 'onnx'

        out_path = 'mymodel'

        self.boring_model.save(out_path)
        converter = Converter(out_format=out_format, model_path=out_path)
        converter.convert()

        self.assertTrue(os.path.exists(converter.out_path))
        

    
    
    def test_convert_tflite_format(self):
        """
        GIVEN a Converter correctly initialized with tflite out_format
        WHEN the convert method is called
        THEN the model should be converted to tflite format
        """ 
        out_format = 'tflite'
        converter = Converter(out_format=out_format, model=self.boring_model)
        converter.convert()
        self.assertTrue(os.path.exists(converter.out_path))

    def test_convert_tflite_format_with_cli(self):
        """
        GIVEN a Converter correctly initialized with tflite out_format and without model
        WHEN the convert method is called with the cli
        THEN the model should be converted to tflite format
        """ 
        out_format = 'tflite'

        out_path = 'mymodel'

        self.boring_model.save(out_path)
        converter = Converter(out_format=out_format, model_path=out_path)
        converter.convert()

        self.assertTrue(os.path.exists(converter.out_path))

    def test_convert_onnx_wrong_name_model(self):
        """
        GIVEN a Converter correctly initialized with onnx out_format
        WHEN the convert method is called with a out_name not ending with .onnx
        THEN the model should anyways save the model with the correct extension
        """ 
        out_format = 'onnx'
        converter = Converter(out_path = 'model', out_format=out_format, model=self.boring_model)
        converter.convert()
        self.assertEqual(converter.out_path, 'model.onnx')

    def test_convert_tflite_wrong_name_model(self):
        """
        GIVEN a Converter correctly initialized with tflite out_format
        WHEN the convert method is called with a out_name not ending with .tflite
        THEN the model should anyways save the model with the correct extension
        """ 
        out_format = 'tflite'
        converter = Converter(out_path = 'model', out_format=out_format, model=self.boring_model)
        converter.convert()
        self.assertEqual(converter.out_path, 'model.tflite')
        

    
        


    



    


"""
Pydantic models for YAKET Trainer.

This package requires the following modules:
    - `pydantic`: https://pydantic-docs.helpmanual.io/
    - `pyyaml`: https://pyyaml.org/wiki/PyYAMLDocumentation
"""



import gc
import importlib
import os
import subprocess as sp
import sys
import time
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import mlflow
import numpy as np
import tensorflow as tf

from yaket.converter.converter import Converter
from yaket.schema.schema import Accelerator, TrainingModel, yaml_to_pydantic
import platform

@dataclass
class Trainer:
    """
    A class to train a model using YAML config file
    
    Args:
        config_params (Dict | str): Dictionary or path to YAML config file
        model (tf.keras.Model): Model to train
        train_dataset (Tuple | tf.data.Dataset): Training dataset 
        val_dataset (Tuple | tf.data.Dataset): Validation dataset
        strategy (Optional[tf.distribute.Strategy]):  Strategy to use for training (default None)
        random_seed (int):  Random seed to use for reproducibility
        custom_modules_path (str|Path): Path to the script with custom modules to use for training (default None)
        validate_yaml (bool):  Whether or not to validate the YAML config file (default True)

    Methods:
        - train(epochs) - Main function to call. Return history trained model
        - convert_model(format_model, opset_onnx, output_path, from_command_line) - Convert the model to a different format. Available formats: onnx, tflite
        - validate_config() - Validate again the configuration file. Usefult to use after chaning the config parameters.
        - clear_ram() - Clear RAM
        - get_free_gpu_idx() - Get the index of the free GPU. It uses nvidia-smi and it works only on Linux with GPU.
        - list_available_tf_modules(options) - List available TensorFlow modules for  Optimizers, Losses, and Metrics
    """

    config_params: Union[Dict, str]
    model: tf.keras.Model
    train_dataset: Union[
        Tuple[np.ndarray, np.ndarray],
        Tuple[np.ndarray, np.ndarray, np.ndarray],
        tf.data.Dataset,
    ]
    val_dataset: Union[
        Tuple[np.ndarray, np.ndarray],
        Tuple[np.ndarray, np.ndarray, np.ndarray],
        tf.data.Dataset,
    ] = None
    strategy: Optional[tf.distribute.Strategy] = None
    random_seed : int = 1234
    validate_yaml: bool = True
    custom_modules_path: Optional[str] = None

    # internals
    _config: TrainingModel = None
    _input_shape: Tuple[int, ...] = None
    _metrics = List[Union[tf.keras.metrics.Metric, Callable]]
    _callbacks: List[tf.keras.callbacks.Callback] = None
    _optimizer: tf.keras.optimizers.Optimizer = None
    _loss: Union[tf.keras.losses.Loss, Callable] = None
    _custom_module: Callable = None
    _history: Dict[str, Any] = None
    _model_checkpoint: Optional[str] = None
    _accelerator: Optional[Accelerator] = None
    _log: bool = False
    _out_path: str = None
    _user_strategy: bool = False
    _trainer_initialized: bool = False
    _sample_weight_mode: Optional[str] = None

    # __post_init__ is called after the __init__ function automatically

    def train(self, epochs: int = None) -> Dict[str, Any]:
        """Train the model. Main function to call.

        Args:
            epochs:  Number of epochs to train. If None, will train with the number of epochs specified in the config file.
        
        Returns:
            History of the training process

        """

        if self.strategy is None and self._accelerator is Accelerator.cpu:
            train_dataset = self._get_x_y_train(self.config.batch_size)
            val_dataset = self._get_x_y_val(self.config.batch_size)
            history = self._train(train_dataset, val_dataset, epochs)
        else:
            strategy = self._get_strategy()
            batch_size = strategy.num_replicas_in_sync * self.config.batch_size
            train_dataset = self._get_x_y_train(batch_size)
            val_dataset = self._get_x_y_val(batch_size)
            with strategy.scope():
                self._clone_model() if not self._user_strategy else None  # Clone model only if not using user strategy (e.g. CustoModule)
                history = self._train(train_dataset, val_dataset, epochs)

        self._save_model()
        self._history = history.history
        self._clean_workspace()
        return history

    def _train(self, train_dataset, val_dataset, epochs: int = None):
        """Train the model"""

        self._compile_model()
        self._autolog()
        history = self.model.fit(
            x=train_dataset,
            y=None,
            epochs=int(self.config.epochs) if epochs is None else epochs,
            validation_data=val_dataset,
            batch_size=None,
            callbacks=self._callbacks,
            steps_per_epoch=int(self.config.steps_per_epoch)
            if self.config.steps_per_epoch is not None
            else None,
            class_weight=None,  # TODO: add class_weight,
            verbose=int(self.config.verbose),
        )
        return history

    


    def convert_model(
        self,
        format_model: str = "onnx",
        opset_onnx: int = 15,
        output_path: str = "model",
        from_command_line: bool = True,
    ):

        """
        Convert the model to a different format. Available formats:
        1. ONNX
        2. TensorFlow Lite

        Args:
            
            format_model:  Format to convert the model to. Available formats: onnx, tflite
            opset_onnx:  Opset to use for the ONNX model (default 15)
            output_path:  Path to save the converted model (default model)
            from_command_line:  Whether or not to convert the model from the command line (default True)

        Raises:
            TypeError: If format_model, output_path, from_command_line, and opset_onnx are not of the correct type.
            NotImplementedError: If from_command_line is True but the model has not been saved.
        """

        if not isinstance(format_model, str):
            raise TypeError("format_model must be a string")
        if not isinstance(output_path, str):
            raise TypeError("output_path must be a string")
        if not isinstance(from_command_line, bool):
            raise TypeError("from_command_line must be a boolean")
        if not isinstance(opset_onnx, int) or (isinstance(opset_onnx, int) and opset_onnx < 1):
            raise TypeError("opset_onnx must be a positive integer")

        if from_command_line and self._out_path is None:
            raise NotImplementedError(
                "You need to have a saved model before converting using command line"
            )
        model_to_convert = self.model if not from_command_line else None
        converter = Converter(
            out_format=format_model.lower(),
            opset_onnx=opset_onnx,
            model_path=self._out_path,
            model=model_to_convert,
            out_path=output_path,
        )
        if converter.convert():
            print(f"Successfully converted to format {format} ")

    def _get_sample_weight_mode(self):
        """Get the sample weight mode from the config file"""

        if self.config.sample_weight_mode is not None:
            return self.config.sample_weight_mode

    def _get_input_shape(self):
        """Get the input shape of input dataset"""

        if isinstance(self.train_dataset, tf.data.Dataset):
            for a in self.train_dataset.take(1):
                self.input_shape = (None, *a[0].shape[1:])
        else:
            self.input_shape = (None, *self.train_dataset[0].shape[1:])
        return self.input_shape

    def _save_model(self):
        """Save the model by loading best checkpoint if available and saving it to mlflow or local path"""

        if self._model_checkpoint is not None:
            self.model.load_weights(self._model_checkpoint)
            if self._log:
                run = mlflow.last_active_run()
                idx = 7  # TODO: check is always the same
                artifact_path = run.info.artifact_uri[idx:]
                self._out_path = artifact_path + f"/best_model"
            else:
                os.makedirs(os.getcwd() + "/models", exist_ok=True)
                t = int(time.time())
                self._out_path = os.getcwd() + f"/models/{t}_best_model"
        else:
            os.makedirs(os.getcwd() + "/models", exist_ok=True)
            t = int(time.time())
            self._out_path = os.getcwd() + f"/models/{t}_best_model"

        try:
            self.model.save(self._out_path)  # if not custom model/layers
        except:
            self.model.save_weights(self._out_path, save_format="tf")

    def _clone_model(self):
        """
        Clone the model so that it works within tf.distribute.Strategy
        It works only for models not using custom objects
        """

        if not self.model._is_graph_network:
            raise Exception(
                "Model must be be a Sequential or Functional API model without custom objects"
            )
        try:
            self.model = tf.keras.models.clone_model(self.model)
        except NotImplementedError:  # there's custom layer
            raise

    def _get_x_y_val(self, batch_size: int) -> tf.data.Dataset:
        """Get the x and y for training based on the format of the dataset"""

        options = tf.data.Options()
        options.experimental_distribute.auto_shard_policy = (
            tf.data.experimental.AutoShardPolicy.OFF
        )

        if self.val_dataset is None:
            return None
        if isinstance(self.val_dataset, tf.data.Dataset):
            return self.val_dataset.with_options(options)
        else:
            val = (
                tf.data.Dataset.from_tensor_slices(self.val_dataset)
                .batch(batch_size)
                .with_options(options)
            )
            return val

    def _get_x_y_train(self, batch_size: int) -> tf.data.Dataset:
        """Get the x and y for training based on the format of the dataset"""

        options = tf.data.Options()
        options.experimental_distribute.auto_shard_policy = (
            tf.data.experimental.AutoShardPolicy.OFF
        )

        if isinstance(self.train_dataset, tf.data.Dataset):
            x = self.train_dataset.with_options(options)
        else:
            x = tf.data.Dataset.from_tensor_slices(self.train_dataset)
            if self.config.shuffle:
                x = x.shuffle(self.train_dataset[0].shape[0])
            x = x.batch(batch_size).prefetch(1).with_options(options)
        return x

    def validate_config(self):
        """Validate again the configuration file. Used after chaning the config parameters"""

        temp_dir = {k: v for k, v in self.config.__dict__.items() if v is not None}

        TrainingModel(**temp_dir)
        self._validate_config_file()

    @property
    def config(self):
        return self._config

    def _compile_model(self) -> None:
        """Compile the model"""

        self._optimizer = self._get_optimizer()
        self._loss = self._get_loss()
        self._metrics = self._get_metrics()

        self.model.compile(
            optimizer=self._optimizer,
            loss=self._loss,
            metrics=self._get_metrics(),
            sample_weight_mode=self._sample_weight_mode,
        )

    def _get_strategy(self) -> tf.distribute.Strategy:
        """Get the strategy for the model"""

        if self.strategy is None:
            if self._accelerator is None:
                return tf.distribute.MirroredStrategy()
            if self._accelerator is Accelerator.gpu:
                index = Trainer.get_free_gpu_idx()
                return tf.distribute.OneDeviceStrategy(f"/gpu:{index}")
            if (
                self._accelerator is Accelerator.cpu
                or self._accelerator is Accelerator.mgpu
            ):
                # If GPUs are not available, it will use CPUs
                return tf.distribute.MirroredStrategy()
            if self._accelerator is Accelerator.tpu:
                # TODO: check configuration for tpu strategy
                return tf.distribute.TPUStrategy()
        else:
            self._user_strategy = True
            return self.strategy

    @staticmethod
    def get_free_gpu_idx() -> int:
        """
        Get the index of the freer GPU

        Raises:
            Exception: If no GPU is available or system is not Linux

        
        Returns:        
            Index of the freer GPU
        """

        if platform.system == "Windows":
            raise Exception("GPU not supported on Windows")
        gpus = tf.config.experimental.list_physical_devices("GPU")
        if len(gpus) == 0:
            raise Exception("No GPU available")

        command = "nvidia-smi --query-gpu=memory.free --format=csv"
        memory_free_info = (
            sp.check_output(command.split()).decode("ascii").split("\n")[:-1][1:]
        )
        memory_free_values = [int(x.split()[0]) for i, x in enumerate(memory_free_info)]
        return int(np.argmin(memory_free_values))

    def _parse_config(self) -> Any:
        """Parse the config file"""

        if isinstance(self.config_params, str):
            return yaml_to_pydantic(self.config_params, self.validate_yaml)
        if isinstance(self.config_params, dict):
            return TrainingModel(**self.config_params)

    def _validate_config_file(self):
        "Validate existence of the loss, optimizer and callbacks defined in the config file"

        try:
            self._get_optimizer()
            self._get_metrics()
            self._get_loss()
        except Exception as e:
            raise ValueError(
                f"You are using a module not defined in either keras or in the custom script\n Details: {e}"
            )

    def _import_custom_model(self, module_name: str):
        """Import the custom model from the custom script"""

        try:
            custom_dirpath = os.path.dirname(module_name)
            sys.path.append(custom_dirpath)
            module = module_name.split("/")[-1].split(".")[0]
            self._custom_module = importlib.import_module(module)
        except Exception as e:
            raise ImportError(f"Error importing {module_name}: {e}")

    def _load_custom_module(
        self, module_name: str, params: Optional[Dict] = None
    ) -> Callable:
        """Load the custom model from the custom script"""
        
        try:
            if params is None:
                return getattr(self._custom_module, module_name)
            else:
                return getattr(self._custom_module, module_name)(**params)
        except Exception as e:
            raise ImportError(f"Error importing {module_name}: It does not exist")

    def _get_optimizer(self) -> tf.keras.optimizers.Optimizer:
        """Get the optimizer from the config file"""
        
        default_value = "not_found"
        opt_pars = dict()
        opt = self.config.optimizer
        if isinstance(opt, List):
            if len(opt) == 1 and isinstance(opt[0], Dict):
                k = list(opt[0].keys())[0]
                v = list(opt[0].values())[0]
                optimizer = getattr(tf.keras.optimizers, k, default_value)
                if optimizer != default_value:
                    return optimizer(**opt_pars)
                else:
                    return self._load_custom_module(optimizer, opt_pars)
            else:
                opt_name = opt[0]
                optimizer = getattr(tf.keras.optimizers, f"{opt_name}", default_value)
                if optimizer == default_value:
                    optimizer = self._load_custom_module(optimizer)
                if isinstance(opt[1], dict):
                    k = list(opt[1].keys())[0]
                    v = list(opt[1].values())[0]
                    scheduler = getattr(tf.keras.optimizers.schedules, k, default_value)
                    if scheduler == default_value:
                        raise ValueError(
                            f"{v} is not a valid scheduler. Only available from keras"
                        )
                    else:
                        opt_pars["learning_rate"] = scheduler(**v)
                        return optimizer(**opt_pars)
        elif isinstance(opt, str):
            optimizer = getattr(
                tf.keras.optimizers, f"{self.config.optimizer}", default_value
            )
            if optimizer != default_value:
                return optimizer(**opt_pars)
            else:
                return self._load_custom_module(optimizer, opt_pars)

    def _get_loss(self) -> Union[tf.keras.losses.Loss, Callable]:
        """Get the loss from the config file"""

        loss_config = self.config.loss
        default_value = "not_found"

        if isinstance(loss_config, str):
            loss = getattr(tf.keras.losses, loss_config, default_value)
            if loss != default_value:
                return loss()
            else:  # it's a custom loss
                return self._load_custom_module(loss_config)
        elif isinstance(loss_config, Dict):
            loss_name = list(loss_config.keys())[0]
            loss_params = list(loss_config.values())[0]
            loss = getattr(tf.keras.losses, loss_name, default_value)
            if loss != default_value:
                return loss(**loss_params)
            else:
                return self._load_custom_module(loss_config, loss_params)

    def _get_callbacks(self) -> List[tf.keras.callbacks.Callback]:
        """Get the callbacks from the config file"""

        if self.config.callbacks is None:
            return None
        callbacks = []
        default_value = "not_found"

        for name_callback in self.config.callbacks:
            if isinstance(name_callback, Dict):
                key = list(name_callback.keys())[0]
                args = list(name_callback.values())[0]

                # Track filepath if it's a ModelCheckpoint
                self._model_checkpoint = (
                    args["filepath"] if key == "ModelCheckpoint" else None
                )

                callback_value = getattr(tf.keras.callbacks, key, default_value)
            else:
                callback_value = getattr(
                    tf.keras.callbacks, name_callback, default_value
                )
                args = None

            if callback_value != default_value:
                callbacks.append(
                    callback_value(**args) if args is not None else callback_value()
                )
            else:
                if args is not None:
                    if args.get("dataset") is not None:
                        dataset = (
                            self.val_dataset
                            if args["dataset"] in ["val", "dev", "validation"]
                            else self.train_dataset
                        )
                        args["dataset"] = dataset
                    callbacks.append(self._load_custom_module(key, args))
                else:
                    callbacks.append(self._load_custom_module(name_callback, args))

        return callbacks

    def _get_metrics(self) -> List[Union[tf.keras.metrics.Metric, Callable]]:
        """Get the metrics"""

        if self.config.metrics is None:
            return None

        list_metrics = []
        default_value = "not_found"
        for metric in self.config.metrics:
            if metric is None:
                continue
            if isinstance(metric, str):
                args = None
                metric_value = getattr(tf.keras.metrics, f"{metric}", default_value)()
            else:
                m, args = list(metric.items())[0]
                metric_value = getattr(tf.keras.metrics, f"{m}", default_value)(**args)

            if metric_value != default_value:
                list_metrics.append(metric_value)
            else:
                list_metrics.append(self._load_custom_module(metric, args))
        return list_metrics

    @staticmethod
    def list_available_tf_modules(option: str) -> List[str]:
        """
        List available optimizers, losses, and metrics in tf.keras

        Args:
            option: The option to list. It can be "optimizers", "losses", or "metrics"
        
        Raises:
            AssertionError: if the option is not one of "optimizers", "losses", or "metrics"
        
        Returns:
            The list of available options from tf.keras
        """

        options_func = {
            "optimizers": tf.keras.optimizers,
            "losses": tf.keras.losses,
            "metrics": tf.keras.metrics,
        }
        assert option in list(options_func.keys())
        modules = [value for value in dir(options_func[option]) if value[0].isupper()]
        return modules

    def _clean_workspace(self):
        """Clean the workspace"""

        tf.keras.backend.clear_session()

    def _autolog(self):
        """Autolog the model using MLFlow"""

        if self.config.autolog:
            mlflow.tensorflow.autolog(log_models=True, disable=False)
            self._log = True

    def _set_randomness(self, random_seed: Optional[int] = None) -> None:
        """Set the randomness"""

        if (not isinstance(random_seed, int)) or (isinstance(random_seed, int) and random_seed < 0):
            raise ValueError("random_seed must be a positive integer")
        if random_seed is not None:
            if tf.__version__ >= "2.9.0":
                tf.random.set_seed(random_seed)
                tf.config.experimental.enable_op_determinism()
            else:
                tf.random.set_seed(random_seed)
                np.random.seed(random_seed)

    def clear_ram(self, clear_model: bool = True):
        """
        Clear ram after deleting model and all datasets saved in the Trainer class

        Args:
            clear_model: If True, the model will be deleted from the Trainer class
        """

        if clear_model:
            del self.model
        if self.train_dataset is not None:
            del self.train_dataset
        if self.val_dataset is not None:
            del self.val_dataset
        gc.collect()

    def summary_model(self):
        """Summary of the model"""
        self.model.summary()

    def __post_init__(self):
        """Initialize the trainer: check inputs, load custom modules, parse configuration file."""

        if not isinstance(self.model, tf.keras.models.Model):
            raise TypeError("model must be keras model")

        if not (isinstance(self.train_dataset, tuple) or isinstance(
            self.train_dataset, tf.data.Dataset
        )):
            raise TypeError("train_dataset must be a tuple or a tf.data.Dataset")
        if not (isinstance(self.val_dataset, tuple) or isinstance(
            self.val_dataset, tf.data.Dataset
        ) or self.val_dataset is None):
            raise TypeError("val_dataset must be a tuple or a tf.data.Dataset")
        if isinstance(self.train_dataset, tuple):
            if len(self.train_dataset) < 2 or len(self.train_dataset) > 3:
                raise ValueError(
                    "train_dataset must be a tuple of (x, y) or (x, y, sample_weight)"
                )
            for val in self.train_dataset:
                if not isinstance(val, np.ndarray):
                    raise TypeError("train_dataset must be a tuple of numpy arrays")

        if isinstance(self.val_dataset, tuple):
            if len(self.val_dataset) < 2 or len(self.val_dataset) > 3:
                raise ValueError(
                    "val_dataset must be a tuple of (x, y) or (x, y, sample_weight)"
                )
            for val in self.val_dataset:
                if not isinstance(val, np.ndarray):
                    raise TypeError("val_dataset must be a tuple of numpy arrays")
        if not (
            isinstance(self.config_params, dict) or isinstance(self.config_params, str)
        ):
            raise TypeError("Config_params must be a dictionary or a file path")
        if isinstance(self.config_params, str) and not os.path.isfile(
            self.config_params
        ):
            raise FileNotFoundError("Config_params must be a valid file path")

        if self.strategy is not None and not isinstance(
            self.strategy, tf.distribute.Strategy
        ):
            raise TypeError("Strategy must be keras strategy object")
        if not isinstance(self.random_seed, int):
            raise TypeError("Random seed must be an integer")
        if not isinstance(self.validate_yaml, bool):
            raise TypeError("Validate yaml must be a boolean value")
        if self.custom_modules_path is not None:
            if not isinstance(self.custom_modules_path, str) or not os.path.isfile(
                self.custom_modules_path
            ):
                raise FileNotFoundError(
                    "Costum modules path must be a valid path string"
                )

        if self.custom_modules_path:
            self._import_custom_model(self.custom_modules_path)
        self._config = self._parse_config()
        self._accelerator = Accelerator[self.config.accelerator]
        self._validate_config_file()

        self._callbacks = self._get_callbacks()
        self._input_shape = self._get_input_shape()
        self._sample_weight_mode = self._get_sample_weight_mode()

        self._trainer_initialized = True



"""
Yaket Trainer.

This package requires the following modules:
    - `tensorflow`: https://www.tensorflow.org/install
    - `mlflow`: https://mlflow.org/docs/latest/index.html
    - `yaket.converter.converter` : Prodive a module to convert models to onnx or tflite
    - `yaket.trainer.schema` : Provide a Pydantic module for the Trainer class
    - `numpy`: https://numpy.org/
"""

from .trainer import Trainer

__version__ = "1.3.6"



from typing import Dict, Optional, Any, Union
import yaml
import os
from pydantic import (
    BaseModel,
    validator,
    Extra,
    PositiveInt,
    conint,
    conlist,
    constr,
)
from enum import Enum, auto
import tensorflow as tf



class Accelerator(Enum):
    """Enum for accelerator types."""

    cpu = auto()
    gpu = auto()
    mgpu = auto()
    tpu = auto()

    @classmethod
    def list(cls):
        return list(map(lambda c: c.name, cls))

    @classmethod
    def list_gpu(cls):
        return list(map(lambda c: c.name if "gpu" in c.name else "", cls))


class TrainingModel(BaseModel, extra=Extra.allow):
    """Pydantic model for YAKET Trainer."""

    autolog: bool = False
    optimizer: Union[
        constr(strict=True), conlist(item_type=Dict, min_items=1, max_items=2)
    ] = "Adam"
    optimizer_params: Optional[Dict[str, Any]] = None  # obsolete
    epochs: PositiveInt = 1
    batch_size: PositiveInt = 1  # if format is numpy
    loss: Union[constr(strict=True), Dict[str, Any]]
    callbacks: Optional[conlist(item_type=Union[str, Dict[str, Any]], min_items=0)]
    metrics: Optional[
        conlist(item_type=Union[str, Dict], min_items=1, unique_items=True)
    ]
    verbose: conint(ge=1, le=2) = 1
    shuffle: bool = True
    accelerator: Optional[constr(strict=True)] = "cpu"
    steps_per_epoch: Optional[PositiveInt] = None
    sample_weight_mode: Optional[constr(strict=True)] = None

    @validator("sample_weight_mode")
    def sample_weight_mode_validator(cls, v):
        if v is not None:
            assert v in ["temporal"]
        return v

    @validator("accelerator")
    def accelerator_validator(cls, v):
        if v is None:
            return None
        if v not in Accelerator.list():
            raise ValueError(
                f"{v} is not a valid accelerator.\nPlease use: {Accelerator.list()}"
            )
        if v in Accelerator.list_gpu() and len(tf.config.list_physical_devices("GPU")) == 0:
            raise ValueError("ERROR: No GPU has been detected. Change accelerator.")
        return v


def yaml_to_pydantic(path: str, validate: bool) -> TrainingModel:
    """
    Load YAML file and convert it to a Pydantic model.

    Args:
        path: Path to YAML file.
        validate: Validate the YAML file.

    Returns:
        Pydantic model for TrainingModel.
    """

    
    if not os.path.exists(path=path):
        raise FileNotFoundError(f"{path} not found")
    if os.path.isfile(path) and path.endswith(".yaml"):
        with open(path, "r") as f:
            data = yaml.safe_load(f)
        return (
            TrainingModel(**data)
            if validate
            else TrainingModel.construct(_fields_set=None, **data)
        )
    else:
        raise ValueError(f"{path} is not a yaml file")
